@@ -48,11 +48,14 @@ SERVICE_RENAMES = {
     'iot1click-projects': 'iot-1click-projects',
     'iotevents-data': 'iot-events-data',
     'iotevents': 'iot-events',
+    'iotwireless': 'iot-wireless',
     'kinesisanalytics': 'kinesis-analytics',
     'kinesisanalyticsv2': 'kinesis-analytics-v2',
     'kinesisvideo': 'kinesis-video',
     'lex-models': 'lex-model-building-service',
+    'lexv2-models': 'lex-models-v2',
     'lex-runtime': 'lex-runtime-service',
+    'lexv2-runtime': 'lex-runtime-v2',
     'logs': 'cloudwatch-logs',
     'machinelearning': 'machine-learning',
     'marketplacecommerceanalytics': 'marketplace-commerce-analytics',
@@ -68,6 +71,7 @@ SERVICE_RENAMES = {
     'secretsmanager': 'secrets-manager',
     'serverlessrepo': 'serverlessapplicationrepository',
     'servicecatalog': 'service-catalog',
+    'servicecatalog-appregistry': 'service-catalog-appregistry',
     'stepfunctions': 'sfn',
     'storagegateway': 'storage-gateway',
 }
@@ -86,7 +90,9 @@ ENDPOINT_PREFIX_OVERRIDE = {
     # iotevents data endpoints.json and service-2.json don't line up.
     'ioteventsdata': 'data.iotevents',
     'iotsecuredtunneling': 'api.tunneling.iot',
+    'iotwireless': 'api.iotwireless',
 }
+
 NOT_SUPPORTED_IN_SDK = [
     'mobileanalytics',
     'transcribestreaming',
