@@ -471,6 +471,7 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         r'AWS4-HMAC-SHA256 '
         r'Credential=\w+/\d+/'
         r'(?P<signing_region>[a-z0-9-]+)/'
+        r'(?P<signing_name>[a-z0-9-]+)/'
     )
 
     def setUp(self):
@@ -489,10 +490,32 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
             auth_header).group('signing_region')
         self.assertEqual(expected_region, actual_region)
 
+    def assert_signing_name(self, request, expected_name):
+        auth_header = request.headers['Authorization'].decode('utf-8')
+        actual_name = self._V4_AUTH_REGEX.match(
+            auth_header).group('signing_name')
+        self.assertEqual(expected_name, actual_name)
+
     def assert_signing_region_in_url(self, url, expected_region):
         qs_components = parse_qs(urlsplit(url).query)
         self.assertIn(expected_region, qs_components['X-Amz-Credential'][0])
 
+    def assert_expected_copy_source_header(self,
+                                           http_stubber, expected_copy_source):
+        request = self.http_stubber.requests[0]
+        self.assertIn('x-amz-copy-source', request.headers)
+        self.assertEqual(
+            request.headers['x-amz-copy-source'], expected_copy_source)
+
+    def add_copy_object_response(self, http_stubber):
+        http_stubber.add_response(
+            body=b'<CopyObjectResult></CopyObjectResult>'
+        )
+
+    def assert_endpoint(self, request, expected_endpoint):
+        actual_endpoint = urlsplit(request.url).netloc
+        self.assertEqual(actual_endpoint, expected_endpoint)
+
     def test_missing_region_in_arn(self):
         accesspoint_arn = (
             'arn:aws:s3::123456789012:accesspoint:myendpoint'
@@ -539,12 +562,24 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         accesspoint_arn = (
             'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint'
         )
-        self.client, _ = self.create_stubbed_s3_client(
+        self.client, http_stubber = self.create_stubbed_s3_client(
             endpoint_url='https://custom.com')
-        with self.assertRaises(
-                botocore.exceptions.
-                UnsupportedS3AccesspointConfigurationError):
-            self.client.list_objects(Bucket=accesspoint_arn)
+        http_stubber.add_response()
+        self.client.list_objects(Bucket=accesspoint_arn)
+        expected_endpoint = 'myendpoint-123456789012.custom.com'
+        self.assert_endpoint(http_stubber.requests[0], expected_endpoint)
+
+    def test_accesspoint_arn_with_custom_endpoint_and_dualstack(self):
+        accesspoint_arn = (
+            'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint'
+        )
+        self.client, http_stubber = self.create_stubbed_s3_client(
+            endpoint_url='https://custom.com',
+            config=Config(s3={'use_dualstack_endpoint': True}))
+        http_stubber.add_response()
+        self.client.list_objects(Bucket=accesspoint_arn)
+        expected_endpoint = 'myendpoint-123456789012.custom.com'
+        self.assert_endpoint(http_stubber.requests[0], expected_endpoint)
 
     def test_accesspoint_arn_with_s3_accelerate(self):
         accesspoint_arn = (
@@ -629,6 +664,179 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
             'get_object', {'Bucket': accesspoint_arn, 'Key': 'mykey'})
         self.assert_signing_region_in_url(url, 'us-east-1')
 
+    def test_copy_source_str_with_accesspoint_arn(self):
+        copy_source = (
+            'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint/'
+            'object/myprefix/myobject'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client()
+        self.add_copy_object_response(self.http_stubber)
+        self.client.copy_object(
+            Bucket='mybucket', Key='mykey', CopySource=copy_source
+        )
+        self.assert_expected_copy_source_header(
+            self.http_stubber,
+            expected_copy_source=(
+                b'arn%3Aaws%3As3%3Aus-west-2%3A123456789012%3Aaccesspoint%3A'
+                b'myendpoint/object/myprefix/myobject'
+            )
+        )
+
+    def test_copy_source_str_with_accesspoint_arn_and_version_id(self):
+        copy_source = (
+            'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint/'
+            'object/myprefix/myobject?versionId=myversionid'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client()
+        self.add_copy_object_response(self.http_stubber)
+        self.client.copy_object(
+            Bucket='mybucket', Key='mykey', CopySource=copy_source
+        )
+        self.assert_expected_copy_source_header(
+            self.http_stubber,
+            expected_copy_source=(
+                b'arn%3Aaws%3As3%3Aus-west-2%3A123456789012%3Aaccesspoint%3A'
+                b'myendpoint/object/myprefix/myobject?versionId=myversionid'
+            )
+        )
+
+    def test_copy_source_dict_with_accesspoint_arn(self):
+        copy_source = {
+            'Bucket':
+                'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint',
+            'Key': 'myprefix/myobject',
+        }
+        self.client, self.http_stubber = self.create_stubbed_s3_client()
+        self.add_copy_object_response(self.http_stubber)
+        self.client.copy_object(
+            Bucket='mybucket', Key='mykey', CopySource=copy_source
+        )
+        self.assert_expected_copy_source_header(
+            self.http_stubber,
+            expected_copy_source=(
+                b'arn%3Aaws%3As3%3Aus-west-2%3A123456789012%3Aaccesspoint%3A'
+                b'myendpoint/object/myprefix/myobject'
+            )
+        )
+
+    def test_copy_source_dict_with_accesspoint_arn_and_version_id(self):
+        copy_source = {
+            'Bucket':
+                'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint',
+            'Key': 'myprefix/myobject',
+            'VersionId': 'myversionid'
+        }
+        self.client, self.http_stubber = self.create_stubbed_s3_client()
+        self.add_copy_object_response(self.http_stubber)
+        self.client.copy_object(
+            Bucket='mybucket', Key='mykey', CopySource=copy_source
+        )
+        self.assert_expected_copy_source_header(
+            self.http_stubber,
+            expected_copy_source=(
+                b'arn%3Aaws%3As3%3Aus-west-2%3A123456789012%3Aaccesspoint%3A'
+                b'myendpoint/object/myprefix/myobject?versionId=myversionid'
+            )
+        )
+
+    def test_basic_outpost_arn(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456:accesspoint:myaccesspoint'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-east-1')
+        self.http_stubber.add_response()
+        self.client.list_objects(Bucket=outpost_arn)
+        request = self.http_stubber.requests[0]
+        self.assert_signing_name(request, 's3-outposts')
+        self.assert_signing_region(request, 'us-west-2')
+        expected_endpoint = (
+            'myaccesspoint-123456789012.op-01234567890123456.'
+            's3-outposts.us-west-2.amazonaws.com'
+        )
+        self.assert_endpoint(request, expected_endpoint)
+
+    def test_basic_outpost_arn(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456:accesspoint:myaccesspoint'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            endpoint_url='https://custom.com',
+            region_name='us-east-1')
+        self.http_stubber.add_response()
+        self.client.list_objects(Bucket=outpost_arn)
+        request = self.http_stubber.requests[0]
+        self.assert_signing_name(request, 's3-outposts')
+        self.assert_signing_region(request, 'us-west-2')
+        expected_endpoint = (
+            'myaccesspoint-123456789012.op-01234567890123456.custom.com'
+        )
+        self.assert_endpoint(request, expected_endpoint)
+
+    def test_outpost_arn_with_s3_accelerate(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456:accesspoint:myaccesspoint'
+        )
+        self.client, _ = self.create_stubbed_s3_client(
+            config=Config(s3={'use_accelerate_endpoint': True}))
+        with self.assertRaises(
+                botocore.exceptions.
+                UnsupportedS3AccesspointConfigurationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
+    def test_outpost_arn_with_s3_dualstack(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456:accesspoint:myaccesspoint'
+        )
+        self.client, _ = self.create_stubbed_s3_client(
+            config=Config(s3={'use_dualstack_endpoint': True}))
+        with self.assertRaises(
+                botocore.exceptions.
+                UnsupportedS3AccesspointConfigurationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
+    def test_incorrect_outpost_format(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost'
+        )
+        with self.assertRaises(botocore.exceptions.ParamValidationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
+    def test_incorrect_outpost_no_accesspoint(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456'
+        )
+        with self.assertRaises(botocore.exceptions.ParamValidationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
+    def test_incorrect_outpost_resource_format(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:myaccesspoint'
+        )
+        with self.assertRaises(botocore.exceptions.ParamValidationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
+    def test_incorrect_outpost_sub_resources(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456:accesspoint:mybucket:object:foo'
+        )
+        with self.assertRaises(botocore.exceptions.ParamValidationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
+    def test_incorrect_outpost_invalid_character(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-0123456.890123456:accesspoint:myaccesspoint'
+        )
+        with self.assertRaises(botocore.exceptions.ParamValidationError):
+            self.client.list_objects(Bucket=outpost_arn)
+
 
 class TestOnlyAsciiCharsAllowed(BaseS3OperationTest):
     def test_validates_non_ascii_chars_trigger_validation_error(self):
@@ -741,6 +949,18 @@ class TestS3SigV4(BaseS3OperationTest):
             self.client.put_object(Bucket='foo', Key='bar', Body='baz')
         self.assertIn('content-md5', self.get_sent_headers())
 
+    def test_content_md5_set_empty_body(self):
+        with self.http_stubber:
+            self.client.put_object(Bucket='foo', Key='bar', Body='')
+        self.assertIn('content-md5', self.get_sent_headers())
+
+    def test_content_md5_set_empty_file(self):
+        with self.http_stubber:
+            with temporary_file('rb') as f:
+                assert f.read() == b''
+                self.client.put_object(Bucket='foo', Key='bar', Body=f)
+        self.assertIn('content-md5', self.get_sent_headers())
+
     def test_content_sha256_set_if_config_value_is_true(self):
         config = Config(signature_version='s3v4', s3={
             'payload_signing_enabled': True
@@ -771,7 +991,7 @@ class TestS3SigV4(BaseS3OperationTest):
 
     def test_content_sha256_set_if_md5_is_unavailable(self):
         with mock.patch('botocore.auth.MD5_AVAILABLE', False):
-            with mock.patch('botocore.handlers.MD5_AVAILABLE', False):
+            with mock.patch('botocore.utils.MD5_AVAILABLE', False):
                 with self.http_stubber:
                     self.client.put_object(Bucket='foo', Key='bar', Body='baz')
         sent_headers = self.get_sent_headers()
@@ -1106,6 +1326,65 @@ class TestGeneratePresigned(BaseS3OperationTest):
             'get_object', {'Bucket': 'mybucket', 'Key': 'mykey'})
         self.assert_is_v2_presigned_url(url)
 
+def test_checksums_included_in_expected_operations():
+    """Validate expected calls include Content-MD5 header"""
+
+    t = S3ChecksumCases(_verify_checksum_in_headers)
+    yield t.case('put_bucket_tagging',
+            {"Bucket": "foo", "Tagging":{"TagSet":[]}})
+    yield t.case('put_bucket_lifecycle',
+            {"Bucket": "foo", "LifecycleConfiguration":{"Rules":[]}})
+    yield t.case('put_bucket_lifecycle_configuration',
+            {"Bucket": "foo", "LifecycleConfiguration":{"Rules":[]}})
+    yield t.case('put_bucket_cors',
+            {"Bucket": "foo", "CORSConfiguration":{"CORSRules": []}})
+    yield t.case('delete_objects',
+            {"Bucket": "foo", "Delete": {"Objects": [{"Key": "bar"}]}})
+    yield t.case('put_bucket_replication',
+            {"Bucket": "foo",
+             "ReplicationConfiguration": {"Role":"", "Rules": []}})
+    yield t.case('put_bucket_acl',
+            {"Bucket": "foo", "AccessControlPolicy":{}})
+    yield t.case('put_bucket_logging',
+            {"Bucket": "foo",
+             "BucketLoggingStatus":{}})
+    yield t.case('put_bucket_notification',
+            {"Bucket": "foo", "NotificationConfiguration":{}})
+    yield t.case('put_bucket_policy',
+            {"Bucket": "foo", "Policy": "<bucket-policy>"})
+    yield t.case('put_bucket_request_payment',
+            {"Bucket": "foo", "RequestPaymentConfiguration":{"Payer": ""}})
+    yield t.case('put_bucket_versioning',
+            {"Bucket": "foo", "VersioningConfiguration":{}})
+    yield t.case('put_bucket_website',
+            {"Bucket": "foo",
+             "WebsiteConfiguration":{}})
+    yield t.case('put_object_acl',
+            {"Bucket": "foo", "Key": "bar", "AccessControlPolicy":{}})
+    yield t.case('put_object_legal_hold',
+            {"Bucket": "foo", "Key": "bar", "LegalHold":{"Status": "ON"}})
+    yield t.case('put_object_retention',
+            {"Bucket": "foo", "Key": "bar",
+             "Retention":{"RetainUntilDate":"2020-11-05"}})
+    yield t.case('put_object_lock_configuration',
+            {"Bucket": "foo", "ObjectLockConfiguration":{}})
+
+
+def _verify_checksum_in_headers(operation, operation_kwargs):
+    environ = {}
+    with mock.patch('os.environ', environ):
+        environ['AWS_ACCESS_KEY_ID'] = 'access_key'
+        environ['AWS_SECRET_ACCESS_KEY'] = 'secret_key'
+        environ['AWS_CONFIG_FILE'] = 'no-exist-foo'
+        session = create_session()
+        session.config_filename = 'no-exist-foo'
+        client = session.create_client('s3')
+        with ClientHTTPStubber(client) as stub:
+            stub.add_response()
+            call = getattr(client, operation)
+            call(**operation_kwargs)
+            assert 'Content-MD5' in stub.requests[-1].headers
+
 
 def test_correct_url_used_for_s3():
     # Test that given various sets of config options and bucket names,
@@ -1267,7 +1546,7 @@ def test_correct_url_used_for_s3():
     yield t.case(
         region='fips-us-gov-west-1', bucket='bucket', key='key',
         signature_version='s3',
-        expected_url='https://bucket.s3-fips-us-gov-west-1.amazonaws.com/key')
+        expected_url='https://bucket.s3-fips.us-gov-west-1.amazonaws.com/key')
 
 
     # Test path style addressing.
@@ -1759,10 +2038,15 @@ def test_correct_url_used_for_s3():
             'https://bucket.s3.unknown.amazonaws.com/key'))
 
 
-class S3AddressingCases(object):
+class BaseTestCase:
     def __init__(self, verify_function):
         self._verify = verify_function
 
+    def case(self, **kwargs):
+        return self._verify, kwargs
+
+
+class S3AddressingCases(BaseTestCase):
     def case(self, region=None, bucket='bucket', key='key',
              s3_config=None, is_secure=True, customer_provided_endpoint=None,
              expected_url=None, signature_version=None):
@@ -1772,6 +2056,11 @@ class S3AddressingCases(object):
         )
 
 
+class S3ChecksumCases(BaseTestCase):
+    def case(self, operation, operation_args):
+        return self._verify, operation, operation_args
+
+
 def _verify_expected_endpoint_url(region, bucket, key, s3_config,
                                   is_secure=True,
                                   customer_provided_endpoint=None,
