@@ -857,3 +857,24 @@ class TestDefaultClientConfig(BaseSessionTest):
         client_config = botocore.config.Config()
         self.session.set_default_client_config(client_config)
         self.assertIs(self.session.get_default_client_config(), client_config)
+
+
+class TestSessionRegionSetup(BaseSessionTest):
+    def test_new_session_with_valid_region(self):
+        s3_client = self.session.create_client('s3', 'us-west-2')
+        self.assertIsInstance(s3_client, client.BaseClient)
+        self.assertEquals(s3_client.meta.region_name, 'us-west-2')
+
+    def test_new_session_with_unknown_region(self):
+        s3_client = self.session.create_client('s3', 'MyCustomRegion1')
+        self.assertIsInstance(s3_client, client.BaseClient)
+        self.assertEquals(s3_client.meta.region_name, 'MyCustomRegion1')
+
+    def test_new_session_with_invalid_region(self):
+        with self.assertRaises(botocore.exceptions.InvalidRegionError):
+            s3_client = self.session.create_client('s3', 'not.a.real#region')
+
+    def test_new_session_with_none_region(self):
+        s3_client = self.session.create_client('s3', region_name=None)
+        self.assertIsInstance(s3_client, client.BaseClient)
+        self.assertTrue(s3_client.meta.region_name is not None)
